import numpy as np

from visions.backends.numpy.array_utils import array_handle_nulls, array_not_empty
from visions.types.string import String


@array_handle_nulls
def _is_string(array: np.ndarray, state: dict):
    if not all(isinstance(v, str) for v in array[0:5]):
        return False
    try:
        return (array.astype(str) == array).all()
    except (TypeError, ValueError):
        return False


@String.contains_op.register
@array_not_empty
def string_contains(array: np.ndarray, state: dict) -> bool:
    if np.issubdtype(array.dtype, np.str_):
        return True

    return _is_string(array, state)



import numpy as np

from visions.backends.numpy import test_utils
from visions.backends.numpy.array_utils import array_handle_nulls, array_not_empty
from visions.types.complex import Complex
from visions.types.float import Float
from visions.types.string import String
from visions.utils.warning_handling import suppress_warnings


def test_string_leading_zeros(array: np.ndarray, coerced_array: np.ndarray):
    return not any(s[0] == "0" for s in array[coerced_array > 1])


@Float.register_relationship(String, np.ndarray)
@array_handle_nulls
def string_is_float(array: np.ndarray, state: dict) -> bool:
    coerced_array = test_utils.option_coercion_evaluator(lambda s: s.astype(np.float_))(
        array
    )

    return (
        coerced_array is not None
        and float_contains(coerced_array, state)
        and test_string_leading_zeros(array, coerced_array)
    )


@Float.register_transformer(String, np.ndarray)
def string_to_float(array: np.array, state: dict) -> np.ndarray:
    return array.astype(np.float_)


@Float.register_relationship(Complex, np.ndarray)
def complex_is_float(array: np.array, state: dict) -> bool:
    return all(np.imag(array) == 0)


@Float.register_transformer(Complex, np.ndarray)
def complex_to_float(array: np.array, state: dict) -> np.ndarray:
    return suppress_warnings(lambda s: s.astype(np.float_))(array)


@Float.contains_op.register
@array_handle_nulls
@array_not_empty
def float_contains(array: np.ndarray, state: dict) -> bool:
    return np.issubdtype(array.dtype, np.floating)



from typing import Dict, List

import pandas as pd
import pandas.api.types as pdt

from visions.backends.pandas.series_utils import (
    series_handle_nulls,
    series_not_empty,
    series_not_sparse,
)
from visions.backends.pandas.test_utils import (
    coercion_map,
    coercion_map_test,
    pandas_version,
)
from visions.backends.python.types.boolean import get_boolean_coercions
from visions.types.boolean import Boolean
from visions.types.object import Object
from visions.types.string import String

hasnan_bool_name = "boolean" if pandas_version[0] >= 1 else "Bool"


string_coercions = get_boolean_coercions("en")


@Boolean.register_relationship(Object, pd.Series)
@series_handle_nulls
def object_is_boolean(series: pd.Series, state: dict) -> bool:
    bool_set = {True, False}
    try:
        ret = all(item in bool_set for item in series.values)
    except (ValueError, TypeError, AttributeError):
        ret = False

    return ret


@Boolean.register_transformer(Object, pd.Series)
def object_to_boolean(series: pd.Series, state: dict) -> pd.Series:
    dtype = hasnan_bool_name if series.hasnans else bool
    return series.astype(dtype)


@Boolean.register_relationship(String, pd.Series)
def string_is_boolean(series: pd.Series, state: dict) -> bool:
    try:
        return coercion_map_test(string_coercions)(series.str.lower(), state)
    except (ValueError, TypeError, AttributeError):
        return False


@Boolean.register_transformer(String, pd.Series)
def string_to_boolean(series: pd.Series, state: dict) -> pd.Series:
    return object_to_boolean(coercion_map(string_coercions)(series.str.lower()), state)


@Boolean.contains_op.register
@series_not_sparse
@series_handle_nulls
@series_not_empty
def boolean_contains(series: pd.Series, state: dict) -> bool:
    return pdt.is_bool_dtype(series) and not pdt.is_categorical_dtype(series)



from datetime import datetime
from typing import Sequence

from visions.backends.python.series_utils import sequence_not_empty
from visions.types.date_time import DateTime
from visions.types.string import String


@DateTime.register_relationship(String, Sequence)
def string_is_datetime(sequence: Sequence, state: dict) -> bool:
    try:
        _ = list(string_to_datetime(sequence, state))
        return True
    except (OverflowError, TypeError, ValueError):
        return False


@DateTime.register_transformer(String, Sequence)
def string_to_datetime(sequence: Sequence, state: dict) -> Sequence:
    """
    Python 3.7+
    return map(datetime.fromisoformat, sequence)
    """
    return tuple(map(lambda s: datetime.strptime(s, "%Y-%m-%d %H:%M:%S"), sequence))


@DateTime.contains_op.register
@sequence_not_empty
def datetime_contains(sequence: Sequence, state: dict) -> bool:
    return all(isinstance(value, datetime) for value in sequence)



from visions.types.boolean import Boolean
from visions.types.categorical import Categorical
from visions.types.complex import Complex
from visions.types.count import Count
from visions.types.date import Date
from visions.types.date_time import DateTime
from visions.types.email_address import EmailAddress
from visions.types.file import File
from visions.types.float import Float
from visions.types.generic import Generic
from visions.types.geometry import Geometry
from visions.types.image import Image
from visions.types.integer import Integer
from visions.types.ip_address import IPAddress
from visions.types.numeric import Numeric
from visions.types.object import Object
from visions.types.ordinal import Ordinal
from visions.types.path import Path
from visions.types.sparse import Sparse
from visions.types.string import String
from visions.types.time import Time
from visions.types.time_delta import TimeDelta
from visions.types.type import VisionsBaseType
from visions.types.url import URL
from visions.types.uuid import UUID

__all__ = [
    "VisionsBaseType",
    "Generic",
    "String",
    "Boolean",
    "Categorical",
    "Complex",
    "Count",
    "Date",
    "DateTime",
    "File",
    "Float",
    "Geometry",
    "Image",
    "Integer",
    "IPAddress",
    "Object",
    "Ordinal",
    "Path",
    "TimeDelta",
    "UUID",
    "URL",
    "Time",
    "EmailAddress",
    "Sparse",
    "Numeric",
]



import pandas as pd
from pandas.api import types as pdt

from visions.backends.pandas.series_utils import series_not_empty
from visions.types.ordinal import Ordinal

# @Ordinal.register_transformer(Categorical, pd.Series)
# def categorical_to_ordinal(series: pd.Series) -> pd.Categorical:
#     return pd.Categorical(
#         series, categories=sorted(series.dropna().unique()), ordered=True
#     )


@Ordinal.contains_op.register
@series_not_empty
def ordinal_contains(series: pd.Series, state: dict) -> bool:
    return pdt.is_categorical_dtype(series) and series.cat.ordered



import logging

logger = logging.getLogger(__name__)


try:
    import pandas as pd

    import visions.backends.pandas
    from visions.backends.pandas.test_utils import pandas_version

    if pandas_version[0] < 1:
        from visions.dtypes.boolean import BoolDtype
    logger.info(f"Pandas backend loaded {pd.__version__}")

except ImportError:
    logger.info("Pandas backend NOT loaded")


try:
    import numpy as np

    import visions.backends.numpy

    logger.info(f"Numpy backend loaded {np.__version__}")
except ImportError:
    logger.info("Numpy backend NOT loaded")


try:
    import pyspark

    import visions.backends.spark

    logger.info(f"Pyspark backend loaded {pyspark.__version__}")
except ImportError:
    logger.info("Pyspark backend NOT loaded")


try:
    import visions.backends.python

    logger.info("Python backend loaded")
except ImportError:
    logger.info("Python backend NOT loaded")



from typing import Any, Sequence

from multimethod import multimethod

from visions.relations import IdentityRelation, TypeRelation
from visions.types.path import Path
from visions.types.type import VisionsBaseType


class File(VisionsBaseType):
    """**File** implementation of :class:`visions.types.type.VisionsBaseType`.
    (i.e. existing path)

    Examples:
        >>> import pathlib
        >>> x = [pathlib.Path('/home/user/file.txt'), pathlib.Path('/home/user/test2.txt')]
        >>> x in visions.File
        True
    """

    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        relations = [IdentityRelation(Path)]
        return relations

    @staticmethod
    @multimethod
    def contains_op(item: Any, state: dict) -> bool:
        pass



from typing import Any, Sequence

from multimethod import multimethod

from visions.relations import IdentityRelation, TypeRelation
from visions.types.object import Object
from visions.types.type import VisionsBaseType


class String(VisionsBaseType):
    """**String** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = ['rubin', 'carter', 'champion']
        >>> x in visions.String
        True
    """

    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        relations = [IdentityRelation(Object)]
        return relations

    @staticmethod
    @multimethod
    def contains_op(item: Any, state: dict) -> bool:
        pass



from pyspark.sql.dataframe import DataFrame

from visions.types.categorical import Categorical


@Categorical.contains_op.register
def categorical_contains(sequence: DataFrame, state: dict) -> bool:
    return False



import visions.backends.numpy.types.boolean
import visions.backends.numpy.types.complex
import visions.backends.numpy.types.date_time
import visions.backends.numpy.types.float
import visions.backends.numpy.types.integer
import visions.backends.numpy.types.object
import visions.backends.numpy.types.string
import visions.backends.numpy.types.time_delta



import pandas as pd

from visions.backends.pandas import test_utils
from visions.backends.pandas.series_utils import (
    isinstance_attrs,
    series_handle_nulls,
    series_not_empty,
)
from visions.backends.shared.parallelization_engines import pandas_apply
from visions.types.email_address import FQDA, EmailAddress, _to_email
from visions.types.string import String


@EmailAddress.register_relationship(String, pd.Series)
def string_is_email(series: pd.Series, state: dict) -> bool:
    def test_email(s):
        return pandas_apply(pandas_apply(s, _to_email), lambda x: x.local and x.fqdn)

    return test_utils.coercion_true_test(test_email)(series)


@EmailAddress.register_transformer(String, pd.Series)
def string_to_email(series: pd.Series, state: dict) -> pd.Series:
    return pandas_apply(series, _to_email)


@EmailAddress.contains_op.register
@series_not_empty
@series_handle_nulls
def email_address_contains(series: pd.Series, state: dict) -> bool:
    return isinstance_attrs(series, FQDA, ["local", "fqdn"])



from typing import Any, Sequence

from multimethod import multimethod

from visions.relations import TypeRelation
from visions.types.type import VisionsBaseType


class Generic(VisionsBaseType):
    """**Generic** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> import numpy as np
        >>> import visions
        >>> x = ['a', 1, np.nan]
        >>> x in visions.Generic
        True
    """

    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        return []

    @staticmethod
    @multimethod
    def contains_op(item: Any, state: dict) -> bool:
        return True



import functools
from typing import Callable, Sequence


def sequence_not_empty(fn: Callable[..., bool]) -> Callable[..., bool]:
    """Decorator to exclude empty series"""

    @functools.wraps(fn)
    def inner(sequence: Sequence, *args, **kwargs) -> bool:
        if not any(True for _ in sequence):
            return False

        return fn(sequence, *args, **kwargs)

    return inner


def sequence_handle_none(fn: Callable[..., bool]) -> Callable[..., bool]:
    """Decorator for nullable series"""

    @functools.wraps(fn)
    def inner(sequence: Sequence, *args, **kwargs) -> bool:
        sequence = tuple(filter(None, sequence))
        return fn(sequence, *args, **kwargs)

    return inner



from urllib.parse import ParseResult, urlparse

import pandas as pd

from visions.backends.pandas.series_utils import (
    isinstance_attrs,
    series_handle_nulls,
    series_not_empty,
)
from visions.backends.shared.parallelization_engines import pandas_apply
from visions.types.string import String
from visions.types.url import URL


@URL.register_relationship(String, pd.Series)
@series_handle_nulls
def string_is_url(series: pd.Series, state: dict) -> bool:
    try:
        return pandas_apply(
            string_to_url(series, state), lambda x: x.netloc and x.scheme
        ).all()
    except AttributeError:
        return False


@URL.register_transformer(String, pd.Series)
def string_to_url(series: pd.Series, state: dict) -> pd.Series:
    return pandas_apply(series, urlparse)


@URL.contains_op.register
@series_handle_nulls
@series_not_empty
def url_contains(series: pd.Series, state: dict) -> bool:
    return isinstance_attrs(series, ParseResult, ["netloc", "scheme"])



from datetime import timedelta
from typing import Sequence

from visions.backends.python.series_utils import sequence_not_empty
from visions.types.time_delta import TimeDelta


@TimeDelta.contains_op.register
@sequence_not_empty
def time_delta_contains(sequence: Sequence, state: dict) -> bool:
    return all(isinstance(value, timedelta) for value in sequence)



import pathlib
from typing import Sequence

from visions.types.path import Path
from visions.types.string import String


@Path.register_relationship(String, Sequence)
def string_is_path(series, state: dict) -> bool:
    try:
        s = string_to_path(series.copy(), state)
        return all(value.is_absolute() for value in s)
    except TypeError:
        return False


@Path.register_transformer(String, Sequence)
def string_to_path(sequence: Sequence, state: dict) -> Sequence:
    s = tuple(map(pathlib.PureWindowsPath, sequence))
    if not all(value.is_absolute() for value in s):
        return tuple(map(pathlib.PurePosixPath, sequence))
    else:
        return s


@Path.contains_op.register
def path_contains(sequence: Sequence, state: dict) -> bool:
    return all(isinstance(x, pathlib.PurePath) and x.is_absolute() for x in sequence)



import visions.backends.python.types.boolean
import visions.backends.python.types.categorical
import visions.backends.python.types.complex
import visions.backends.python.types.count
import visions.backends.python.types.date
import visions.backends.python.types.date_time
import visions.backends.python.types.email_address
import visions.backends.python.types.file
import visions.backends.python.types.float
import visions.backends.python.types.geometry
import visions.backends.python.types.image
import visions.backends.python.types.integer
import visions.backends.python.types.ip_address
import visions.backends.python.types.numeric
import visions.backends.python.types.object
import visions.backends.python.types.ordinal
import visions.backends.python.types.path
import visions.backends.python.types.string
import visions.backends.python.types.time
import visions.backends.python.types.time_delta
import visions.backends.python.types.url
import visions.backends.python.types.uuid



from datetime import date, time

import pandas as pd

from visions.backends.pandas.series_utils import (
    class_name_attrs,
    series_handle_nulls,
    series_not_empty,
)
from visions.types.date import Date
from visions.types.date_time import DateTime


@Date.register_relationship(DateTime, pd.Series)
@series_handle_nulls
def datetime_is_date(series: pd.Series, state: dict) -> bool:
    dtseries = series.dt.time
    value = time(0, 0)
    return all(v == value for v in dtseries)


@Date.register_transformer(DateTime, pd.Series)
def datetime_to_date(series: pd.Series, state: dict) -> pd.Series:
    return series.dt.date


@Date.contains_op.register
@series_handle_nulls
@series_not_empty
def date_contains(series: pd.Series, state: dict) -> bool:
    return class_name_attrs(series, date, ["year", "month", "day"])



""" Utilities suite for visions """

# from visions.utils.images import image_utils
from visions.utils.monkeypatches import imghdr_patch, pathlib_patch
from visions.utils.profiling import profile_type
from visions.utils.warning_handling import suppress_warnings

__all__ = [
    "profile_type",
    "suppress_warnings",
]



from functools import partial

import pandas as pd
from pandas.api import types as pdt

from visions.backends.pandas import test_utils
from visions.backends.pandas.series_utils import (
    series_handle_nulls,
    series_not_empty,
    series_not_sparse,
)
from visions.types import DateTime, String


@DateTime.register_relationship(String, pd.Series)
@series_handle_nulls
def string_is_datetime(series: pd.Series, state: dict) -> bool:
    exceptions = [OverflowError, TypeError]

    coerced_series = test_utils.option_coercion_evaluator(
        partial(string_to_datetime, state=state), exceptions
    )(series)
    if coerced_series is None:
        return False
    else:
        return not coerced_series.dropna().empty


@DateTime.register_transformer(String, pd.Series)
def string_to_datetime(series: pd.Series, state: dict) -> pd.Series:
    return pd.to_datetime(series)


@DateTime.contains_op.register
@series_not_sparse
@series_handle_nulls
@series_not_empty
def datetime_contains(series: pd.Series, state: dict) -> bool:
    return pdt.is_datetime64_any_dtype(series)



from typing import Any, Sequence

from multimethod import multimethod

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.object import Object
from visions.types.string import String
from visions.types.type import VisionsBaseType


class UUID(VisionsBaseType):
    """**UUID** implementation of :class:`visions.types.type.VisionsBaseType`.

    References:
        UUID specification in RFC4122:
        https://tools.ietf.org/html/rfc4122#section-3

        Python standard library:
        https://docs.python.org/3/library/uuid.html

    Examples:
        >>> import uuid
        >>> import visions
        >>> uuids = ['0b8a22ca-80ad-4df5-85ac-fa49c44b7ede', 'aaa381d6-8442-4f63-88c8-7c900e9a23c6']
        >>> x = [uuid.UUID(uuid_str) for uuid_str in uuids]
        >>> x in visions.UUID
        True
    """

    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        relations = [
            IdentityRelation(Object),
            InferenceRelation(String),
        ]
        return relations

    @staticmethod
    @multimethod
    def contains_op(item: Any, state: dict) -> bool:
        pass



from pathlib import Path
from typing import Union

import networkx as nx


def output_graph(
    G: nx.DiGraph, file_name: Union[Path, str], sort: bool = True, file_format=None
) -> None:
    """Output a graph to a file, either as image or as dot file.

    Args:
        G: the DiGraph to write or plot
        file_name: the file name to write to.
        sort: create a copy of the graph with sorted keys
        file_format: graphviz output format, if None, the file_name extension is used as format
            https://graphviz.org/doc/info/output.html

    Returns:
        Nothing

    Raises:
        ValueError when the file_name does not end on .svg, .png or .dot
    """

    if sort:
        # Create ordered graph for deterministic image outputs
        G_sorted = nx.DiGraph()
        G_sorted.graph["node"] = {"shape": "box", "color": "red"}
        G_sorted.add_nodes_from(sorted(G.nodes, key=lambda x: str(x)))

        style = nx.get_edge_attributes(G, "style")
        for edge in sorted(G.edges, key=lambda x: (str(x[0]), str(x[1]))):
            G_sorted.add_edge(*edge, style=style.get(edge))
        G = G_sorted

    p = nx.drawing.nx_pydot.to_pydot(G)
    if not isinstance(file_name, Path):
        file_name = Path(file_name)

    if file_format is None:
        file_format = file_name.suffix[1:].lower()

    try:
        p.write(file_name, format=file_format)
    except AssertionError:
        raise ValueError(
            "Could not write file. Please make sure that the format is accepted by pydot."
        )
